Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Add copy buttons to all
 blocks
(function() {
function addCopyButtons() {
document.querySelectorAll('pre code').forEach(function(codeBlock) {
if (codeBlock.parentElement.hasAttribute('data-copy-added')) return;
codeBlock.parentElement.setAttribute('data-copy-added', 'true');
var btn = document.createElement('button');
btn.textContent = 'Copy';
btn.style.cssText = 'position:absolute;top:4px;right:4px;padding:2px 8px;font-size:11px;background:#4ecdc4;border:none;border-radius:4px;color:#1a1a2e;cursor:pointer;opacity:0.7;transition:opacity 0.2s;';
btn.onmouseover = function() { this.style.opacity = '1'; };
btn.onmouseout = function() { this.style.opacity = '0.7'; };
btn.onclick = function() {
navigator.clipboard.writeText(codeBlock.textContent).then(function() {
btn.textContent = 'Copied!';
setTimeout(function() { btn.textContent = 'Copy'; }, 1500);
});
};
codeBlock.parentElement.style.position = 'relative';
codeBlock.parentElement.appendChild(btn);
});
}
addCopyButtons();
// Re-run on dynamic content
var observer = new MutationObserver(addCopyButtons);
observer.observe(document.body, { childList: true, subtree: true });
})();
}
} catch(__e) { console.warn('[Userscript:Add Copy Buttons to Code Blocks]', __e); }
})();
(function(){
try {
var __m = "github.com";
var __re = new RegExp('^' + "github\\.com" + '
GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Force GitHub README to respect dark mode (function() { var style = document.createElement('style'); style.textContent = ' .markdown-body { color-scheme: dark light; } .markdown-body pre { background: #161b22 !important; } .markdown-body code { background: rgba(110, 118, 129, 0.4) !important; } .markdown-body table th, .markdown-body table td { border-color: #30363d !important; } .markdown-body img { background: #0d1117; } .markdown-body blockquote { border-left-color: #8b949e; } .markdown-body hr { border-color: #30363d; } '; document.head.appendChild(style); })(); } } catch(__e) { console.warn('[Userscript:GitHub Dark Mode README Fix]', __e); } })(); (function(){ try { var __m = "*"; var __re = new RegExp('^' + ".*" + ' GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Highlight search terms from Google/DuckDuckGo/Bing referrer (function() { var ref = document.referrer; var terms = []; if (ref.includes('google.com') || ref.includes('duckduckgo.com') || ref.includes('bing.com')) { var url = new URL(ref); var q = url.searchParams.get('q') || url.searchParams.get('p'); if (q) { terms = q.split(/\s+/).filter(function(t) { return t.length > 2; }); } } if (terms.length === 0) return; var style = document.createElement('style'); style.textContent = '.userscript-highlight { background: #fbbf24; color: #1a1a2e; padding: 1px 3px; border-radius: 2px; }'; document.head.appendChild(style); function highlight(node) { if (node.nodeType === 3) { // text node var text = node.textContent; var found = false; terms.forEach(function(term) { var regex = new RegExp('(' + term.replace(/[.*+?^${}()|[\]\\]/g, '\\') + ')', 'gi'); if (regex.test(text)) { found = true; var frag = document.createDocumentFragment(); var parts = text.split(regex); parts.forEach(function(part, i) { if (i % 2 === 0) { frag.appendChild(document.createTextNode(part)); } else { var span = document.createElement('span'); span.className = 'userscript-highlight'; span.textContent = part; frag.appendChild(span); } }); node.parentNode.replaceChild(frag, node); } }); } else if (node.nodeType === 1 && node.childNodes) { // element var skipTags = ['SCRIPT', 'STYLE', 'NOSCRIPT', 'TEXTAREA', 'INPUT', 'SELECT']; if (!skipTags.includes(node.tagName)) { Array.from(node.childNodes).forEach(highlight); } } } highlight(document.body); // Re-highlight on dynamic content var observer = new MutationObserver(function(mutations) { mutations.forEach(function(m) { m.addedNodes.forEach(function(node) { if (node.nodeType === 1 || node.nodeType === 3) highlight(node); }); }); }); observer.observe(document.body, { childList: true, subtree: true }); })(); } } catch(__e) { console.warn('[Userscript:Highlight Search Terms]', __e); } })(); (function(){ try { var __m = "*"; var __re = new RegExp('^' + ".*" + ' GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Strip utm_, fbclid, gclid, etc. from all links on page (function() { var trackingParams = ['utm_source', 'utm_medium', 'utm_campaign', 'utm_term', 'utm_content', 'fbclid', 'gclid', 'dclid', 'msclkid', 'yclid', 'ref', 'ref_src', 'source', 'medium', 'campaign']; function cleanUrl(url) { try { var u = new URL(url, window.location.origin); var changed = false; trackingParams.forEach(function(p) { if (u.searchParams.has(p)) { u.searchParams.delete(p); changed = true; } }); return changed ? u.toString() : url; } catch (e) { return url; } } function cleanLinks() { document.querySelectorAll('a[href]').forEach(function(a) { var clean = cleanUrl(a.href); if (clean !== a.href) a.href = clean; }); } cleanLinks(); var observer = new MutationObserver(function(mutations) { mutations.forEach(function(m) { m.addedNodes.forEach(function(node) { if (node.nodeType === 1) { if (node.tagName === 'A') cleanLinks(); node.querySelectorAll('a[href]').forEach(function(a) { var clean = cleanUrl(a.href); if (clean !== a.href) a.href = clean; }); } }); }); }); observer.observe(document.body, { childList: true, subtree: true }); })(); } } catch(__e) { console.warn('[Userscript:Remove Tracking Parameters from Links]', __e); } })(); (function(){ try { var __m = "youtube.com"; var __re = new RegExp('^' + "youtube\\.com" + ' GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Auto-enable theater mode on YouTube (function() { function tryTheater() { var btn = document.querySelector('button[aria-label="Theater mode"], ytd-player #player button[title="Theater mode"]'); if (btn && !btn.classList.contains('activated')) { btn.click(); } } // Try immediately tryTheater(); // Try after navigation (SPA) var lastUrl = location.href; setInterval(function() { if (location.href !== lastUrl) { lastUrl = location.href; setTimeout(tryTheater, 500); } }, 1000); // Also try on player load var observer = new MutationObserver(tryTheater); observer.observe(document.body, { childList: true, subtree: true }); })(); } } catch(__e) { console.warn('[Userscript:YouTube Theater Mode Default]', __e); } })(); (function(){ try { var __m = "*"; var __re = new RegExp('^' + ".*" + ' GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Remove or un-stick sticky/fixed headers that block content (function() { function unstick() { document.querySelectorAll('header, nav, [role="banner"], .header, .navbar, .sticky, .fixed-top, [style*="position: fixed"], [style*="position:sticky"]').forEach(function(el) { if (el.style.position === 'fixed' || el.style.position === 'sticky' || getComputedStyle(el).position === 'fixed' || getComputedStyle(el).position === 'sticky') { el.style.position = 'static'; el.style.top = 'auto'; el.style.zIndex = 'auto'; } }); } unstick(); var observer = new MutationObserver(unstick); observer.observe(document.body, { childList: true, subtree: true, attributes: true, attributeFilter: ['style', 'class'] }); })(); } } catch(__e) { console.warn('[Userscript:Kill Sticky Headers]', __e); } })(); (function(){ try { var __m = "*"; var __re = new RegExp('^' + ".*" + ' GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages

, 'i'); if (__m === '*' || __re.test(location.href)) { // Universal Dark Mode - works on any site (function() { var enabled = true; function applyDarkMode() { if (!enabled) return; // Create style element if it doesn't exist var style = document.getElementById('universal-dark-mode-style'); if (!style) { style = document.createElement('style'); style.id = 'universal-dark-mode-style'; document.head.appendChild(style); } // Dark mode CSS - inverts colors but preserves images/video style.textContent = ' /* Invert everything except media */ html { filter: invert(1) hue-rotate(180deg) !important; background: #1a1a2e !important; } /* Restore images, videos, iframes, canvas */ img, video, iframe, canvas, svg, picture, [style*="background-image"] { filter: invert(1) hue-rotate(180deg) !important; } /* Preserve specific elements that should not be inverted */ .no-dark-mode, .no-dark-mode *, [data-theme="light"], [data-theme="light"], .ace_editor, .ace_editor *, .CodeMirror, .CodeMirror *, .monaco-editor, .monaco-editor *, .markdown-body pre, .markdown-body pre *, .highlight, .highlight *, pre code, pre code * { filter: none !important; } /* Fix common UI elements */ .modal, .popup, .dropdown-menu, .tooltip, .popover { filter: invert(1) hue-rotate(180deg) !important; background: #2d2d44 !important; border-color: #444 !important; } /* Scrollbars */ ::-webkit-scrollbar { background: #1a1a2e !important; } ::-webkit-scrollbar-thumb { background: #444 !important; } ::-webkit-scrollbar-thumb:hover { background: #555 !important; } /* Selection */ ::selection { background: #4ecdc4 !important; color: #1a1a2e !important; } ::-moz-selection { background: #4ecdc4 !important; color: #1a1a2e !important; } '; } function removeDarkMode() { var style = document.getElementById('universal-dark-mode-style'); if (style) style.remove(); } // Toggle with Alt+Shift+D document.addEventListener('keydown', function(e) { if (e.altKey && e.shiftKey && e.key === 'D') { e.preventDefault(); enabled = !enabled; if (enabled) { applyDarkMode(); console.log('[Universal Dark Mode] Enabled'); } else { removeDarkMode(); console.log('[Universal Dark Mode] Disabled'); } } }); // Apply on load applyDarkMode(); // Re-apply on dynamic content var observer = new MutationObserver(function(mutations) { if (enabled && !document.getElementById('universal-dark-mode-style')) { applyDarkMode(); } }); observer.observe(document.head, { childList: true }); console.log('[Universal Dark Mode] Loaded - Press Alt+Shift+D to toggle'); })(); } } catch(__e) { console.warn('[Userscript:Universal Dark Mode]', __e); } })(); })(); GitHub - kanishk393/gpu-kernel-agent-benchmark: Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM · GitHub
Skip to content

Repository files navigation

GPU Kernel Code Generation Benchmark — Multi-Agent vs Zero-Shot LLMs

A domain-specific coding benchmark for CUDA/C++ GPU kernel programming, plus a 3-agent ReAct framework that measurably beats zero-shot generation — evaluated on 25 real-world tasks with an extended CodeBLEU metric and unbiased Pass@1.

PythonMulti-Agent SystemsReActCUDALLM EvaluationCodeBLEUNVIDIA NIMBenchmarking


Why this project

Writing correct GPU kernels is one of the hardest code-generation tasks for LLMs: it demands thread-hierarchy reasoning, memory-layout awareness, synchronization correctness, and CUDA API fluency. It's a perfect stress test for whether agentic frameworks actually improve code quality over plain zero-shot prompting — not just for toy problems, but on 25 tasks collected from 13 real production repositories.

This project answers three questions with data, not vibes:

  1. How well do small/medium open-weight models generate CUDA kernels zero-shot?
  2. Does a multi-agent ReAct framework (Coder → Reviewer → Refiner) improve their scores?
  3. Can a domain-weighted evaluation metric (extended CodeBLEU) capture what makes GPU code correct?

Key Results

ConfigurationPass@1CodeBLEUΔ Pass@1 (agent vs zero-shot)
Bielik-11B (zero-shot)0.8880.3998
Bielik-11B (agent)0.9600.3397+0.072
Nemotron-Nano-9B (zero-shot)0.1560.2548
Nemotron-Nano-9B (agent)0.4000.2269+0.244 (2.6×)
GPT-OSS-20B (zero-shot)0.6520.2951
GPT-OSS-20B (agent)0.6400.2461−0.012 (stable)

The agent harness improves Pass@1 for 2 of 3 models — the Reviewer catches missing __syncthreads, unbalanced memory usage, and wrong indexing; the Refiner fixes them. CodeBLEU drops slightly because agent output is longer and structurally complete, reducing n-gram overlap with compact reference snippets — while actually improving correctness. This gap between surface similarity and functional correctness is exactly why we report both.


Architecture

Three specialized agents collaborate in a Thought → Action → Observation loop, following the multi-agent cross-team collaboration pattern from arXiv:2408.08927:

┌─────────────────────────────────────────────────────────────────────┐
│ Iteration 1 │
│ CoderAgent Thought : analyse task (kernel, memory, sync) │
│ Action : generate initial CUDA code │
│ Observation: code → ReviewerAgent │
├─────────────────────────────────────────────────────────────────────┤
│ Iterations 2–3 │
│ ReviewerAgent Thought : check against 7-point CUDA correctness list │
│ Action : structured JSON review │
│ Observation: review → RefinerAgent │
│ │
│ RefinerAgent Thought : plan fixes for every reported issue │
│ Action : rewrite code addressing all issues │
│ Observation: improved code → next iteration / output │
├─────────────────────────────────────────────────────────────────────┤
│ Early stop: quality = "excellent" AND zero critical issues │
│ AND zero missing components │
└─────────────────────────────────────────────────────────────────────┘

Every step records {iteration, agent, thought, action, observation} — full ReAct traces are saved to results/ for inspection and audit.

Design Decisions

  • Multi-agent, not single-agent. A single "fix your code" loop collapses into echo-chambering. Separating generation, critique, and refinement into distinct agents with different system prompts creates real adversarial pressure — the Reviewer has no incentive to agree with the Coder.
  • Structured JSON reviews. The Reviewer emits a 7-point checklist verdict (critical / major / minor, missing components) instead of free text. This makes refinement deterministic, debuggable, and measurable.
  • Heuristic validity checks instead of nvcc. Compiling 750 generated kernels (25 tasks × 10 samples × 3 models × 2 modes) would need GPU toolchains on every runner. The 6-criterion validator (kernel qualifier, blockIdx+threadIdx, balanced braces, minimum size, no stubs, control flow) is compiler-free and portable — with documented tradeoffs.
  • Domain-weighted CodeBLEU. Standard CodeBLEU treats all tokens equally. GPU code has load-bearing keywords (__syncthreads, __shared__, atomicAdd). We reweighted 42 CUDA keywords at 2.0–5.0× and added a missing-keyword penalty so the metric actually rewards kernel semantics, not just n-gram overlap. Sanity-checked: compute_codebleu(ref, ref) = 1.0000 on all 25 samples.
  • Unbiased Pass@1. Uses the Chen et al. (2021) estimator Pass@1 = c/n with n=10 generations per task — no optimistic bias from sampling without replacement.
  • One API key per model, per mode. Six isolated NIM keys with an RPM limiter so zero-shot and agent phases never contend for rate limits — results stay comparable.
  • Parallel execution. All six phases run in independent screen sessions (launch_all.sh) — a 750-generation benchmark completes in hours, not days.

Benchmark Dataset

25 tasks from 13 real GitHub repositories across five GPU programming categories:

CategoryReposTasks
Core CUDANVIDIA/cuda-samples, moderngpu/moderngpu9
Image Processingopencv/opencv, NaderAlAwar/image-processing-cuda, NajatN/Parallel-Image-Processing-CUDA-, rpgolshan/CUDA-image-processing6
Deep Learninga-hamdi/GPU, SartajBhuvaji/Cuda5
Physics / SimulationBaey/N-Body-CUDA, niteya-shah/Fluid-Simulation-CUDA, vlvovch/lennard-jones-cuda3
Sparse Matrixshreyansh26/SparseMatrix-Computation-CUDA, bsampson1/SpMV-CUDA2

Difficulty distribution: 4 easy · 9 medium · 12 hard

Each sample pairs a natural-language prompt (used verbatim for zero-shot generation) with the ground-truth reference_code from its source repository. Dataset sources are documented per-repo with URLs in data/.

Metrics

Extended CodeBLEU

CodeBLEU = 0.30 × N-gram + 0.40 × Keyword + 0.15 × Dataflow + 0.15 × Syntax

The keyword component (40% weight) uses 42 CUDA-specific keywords:

CategoryKeywordsWeight
Kernel qualifiers__global__, __device__, __shared__4.0–5.0
Thread indexingthreadIdx, blockIdx, blockDim, gridDim3.5–4.5
Synchronisation__syncthreads, __syncwarp3.5–4.5
Memory managementcudaMalloc, cudaFree, cudaMemcpy3.5
Atomic operationsatomicAdd, atomicCAS, etc.3.0–3.5
Warp primitives__shfl_down_sync, __shfl_xor_sync4.0
Kernel launch<<<, >>>4.0

A missing-keyword penalty of −0.10 per absent critical keyword (max −0.50) is applied. Sanity check:compute_codebleu(reference, reference) = 1.0000 for all 25 samples.

Pass@1

Unbiased estimator from Chen et al. (2021): Pass@1 = c / n with n = 10 generations per task, where a generation is valid if it passes all 6 heuristic validity criteria.

Models Evaluated

All served via NVIDIA NIM, one dedicated API key per model:

ModelNIM IdentifierRoles
Bielik 11Bspeakleash/bielik-11b-v2.6-instructZero-shot + Agent
Nemotron Nano 9Bnvidia/nvidia-nemotron-nano-9b-v2Zero-shot + Agent
GPT-OSS 20Bopenai/gpt-oss-20bZero-shot + Agent

Findings

  • Bielik-11B is the strongest model: 0.888 Pass@1 zero-shot → 0.960 with the agent.
  • The agent delivers its largest gains on the weakest model — Nemotron-Nano-9B nearly triples (0.156 → 0.400). Critique-and-refine is a force-multiplier where raw capability is thin.
  • GPT-OSS-20B is stable — the agent neither helps nor hurts significantly (0.652 → 0.640), suggesting its zero-shot output already saturates the validator.
  • CodeBLEU and Pass@1 disagree (agent CodeBLEU ↓ while Pass@1 ↑) — a concrete argument for evaluating generated code by behavior, not just token overlap.

Project Structure

gpu-kernel-agent-benchmark/
├── agent/
│ └── multi_agent_system.py # ReAct: CoderAgent, ReviewerAgent, RefinerAgent
├── benchmark/
│ ├── phase1_bielik_zeroshot.py # Phase 1 — Bielik zero-shot
│ ├── phase2_nemotron_zeroshot.py# Phase 2 — Nemotron zero-shot
│ ├── phase3_gptoss_zeroshot.py # Phase 3 — GPT-OSS zero-shot
│ ├── phase4_bielik_agent.py # Phase 4 — Bielik agent
│ ├── phase5_nemotron_agent.py # Phase 5 — Nemotron agent
│ ├── phase6_gptoss_agent.py # Phase 6 — GPT-OSS agent
│ └── run_benchmark.py # Parallel runner (all 6 phases)
├── data/
│ ├── benchmark_data.py # 25 tasks + metadata
│ ├── samples_11_15.py # Deep learning + image tasks
│ ├── samples_16_20.py # Physics + sparse matrix tasks
│ └── samples_21_25.py # Image processing + simulation tasks
├── evaluation/
│ ├── extended_codebleu.py # Extended CodeBLEU metric
│ └── model_evaluator.py # NIM API + Pass@k + RPM limiter
├── validation/
│ └── basic_validator.py # Heuristic CUDA validator
├── results/ # Full ReAct traces + scores (48 JSON files)
├── demo.py # Self-contained demo (no API keys)
├── launch_all.sh # Launch all 6 screen sessions
├── requirements.txt
├── .env.example
└── README.md

Getting Started

# 1. Install
pip install -r requirements.txt
python -c "import nltk; nltk.download('punkt'); nltk.download('punkt_tab')"# 2. Configure NIM keys
cp .env.example .env # add your 4 NIM API keys# 3. Run the no-API-key demo (self-contained, deterministic)
python demo.py
# 4. Full benchmark — 6 parallel screen sessions
bash launch_all.sh
# 5. Monitor
tail -f results/phase4_bielik_agent_*.log
screen -r p1_bielik_zs

.env format

NIM_BASE_URL=https://integrate.api.nvidia.com/v1
RPM=40
NIM_KEY_NEMOTRON=nvapi-... # Key for Bielik
NIM_KEY_MINISTRAL=nvapi-... # Key for Nemotron
NIM_KEY_MISTRAL=nvapi-... # Key for GPT-OSS
NIM_KEY_AGENT=nvapi-... # Key for agent phases

References

License

MIT — see LICENSE.

About

Multi-agent ReAct framework for CUDA code generation — extended CodeBLEU + unbiased Pass@1 across 25 GPU-kernel tasks from 13 real repos, 3 LLMs via NVIDIA NIM

Topics

Resources

Stars

0 stars

Watchers

0 watching

Forks

Releases

Packages

Contributors

Languages